/********************************************************************
* Description: dualbipod, derived from trivkins.c
*   Kinematics for a dual bipod machine
*   1st bipod x-y plane, 2nd a-b plane
*
*   Derived from a work by Fred Proctor & Will Shackleford
*   20.05.2009: added x00, y00 to enable offsets at reference point
*   04.03.2010: changed to joint0/1 & 3/4
*
* Author:
* License: GPL Version 2
* System: Linux
*    
* Copyright (c) 2004 All rights reserved.
*
* Last change:
* $Revision: 5 $
* $Author: GerhardGleixner@web.de $
* $Date: 2010-02-21 22:39:46 +0100 (So, 21 Feb 2010) $
********************************************************************/

#include "kinematics.h"		/* these decls */
#include "rtapi_math.h"

#include "hal.h"

#include "posemath.h"
#include "rtapi.h"


struct pars {
    hal_float_t JX, JY, DX, DY;
};

struct pars * parMain = 0;

int kinematicsForward(const double *joints,

		      EmcPose * pos,
		      const KINEMATICS_FORWARD_FLAGS * fflags,
		      KINEMATICS_INVERSE_FLAGS * iflags)
{

	double j1, j2, x, y, c; 
	double jx = parMain -> JX; 
	double dx = parMain -> DX;
	double dy = parMain -> DY; 

	j1 = joints[0];
	j2 = joints[1];
	
	x = (j2*j2-j1*j1+2*jx*dx-jx*jx)/(4*dx-2*jx);
	c = j1*j1-(x-dx)*(x-dx)-dy*dy;
	if(c<-1) return -1;
	y = 1+2*sqrt(c+1);
	
    pos->tran.x = x; 
    pos->tran.y = y; 
	
    pos->tran.z = joints[2];


	
    return 0;
}

int kinematicsInverse(const EmcPose * pos,
		      double *joints,
		      const KINEMATICS_INVERSE_FLAGS * iflags,
		      KINEMATICS_FORWARD_FLAGS * fflags)
{
	double x = pos->tran.x;
	double y = pos->tran.y;
	double jx = parMain -> JX; 
	double jy = parMain -> JY;
	double dx = parMain -> DX;
	double dy = parMain -> DY; 

	joints[0] = sqrt((x-dx)*(x-dx)+(y-dy)*(y-dy));
	joints[1] = sqrt((jx-x-dx)*(jx-x-dx)+(y-dy)*(y-dy)) - jx; //jx is the initial length
	joints[2] = sqrt((x-dx)*(x-dx)+(jy-y-dy)*(jy-y-dy)) - jy; //jy is the initial length
	joints[3] = sqrt((jx-x-dx)*(jx-x-dx)+(jy-y-dy)*(jy-y-dy)) - sqrt(jx*jx+jy*jy); // initial length
	
    joints[4] = pos->tran.z;

    return 0;
}

/* implemented for these kinematics as giving joints preference */
/* when is this fct called ???? seems never */
int kinematicsHome(EmcPose * world,
		   double *joint,
		   KINEMATICS_FORWARD_FLAGS * fflags,
		   KINEMATICS_INVERSE_FLAGS * iflags)
{
    *fflags = 0;
    *iflags = 0;


    return kinematicsForward(joint, world, fflags, iflags);
}

KINEMATICS_TYPE kinematicsType()
{
    return KINEMATICS_BOTH;
}

#ifdef RTAPI
#include "rtapi.h"		/* RTAPI realtime OS API */
#include "rtapi_app.h"		/* RTAPI realtime module decls */
#include "hal.h"

#define HAL_SUCCESS 0

EXPORT_SYMBOL(kinematicsType);
EXPORT_SYMBOL(kinematicsForward);
EXPORT_SYMBOL(kinematicsInverse);
EXPORT_SYMBOL(kinematicsHome);
MODULE_LICENSE("GPL");

int comp_id;
int rtapi_app_main(void) 
{
  //struct hal_dat* par;
  int res = 0;

   comp_id = hal_init("quadspeed");
   if(comp_id < 0) return comp_id;

   parMain = hal_malloc(sizeof(*parMain));
   if(!parMain) goto error;

   if((res = hal_param_float_new("quadspeed.setJX", HAL_RW, &parMain->JX, comp_id)) != HAL_SUCCESS) goto error;
   if((res = hal_param_float_new("quadspeed.setJY", HAL_RW, &parMain->JY, comp_id)) != HAL_SUCCESS) goto error;
   if((res = hal_param_float_new("quadspeed.setDX", HAL_RW, &parMain->DX, comp_id)) != HAL_SUCCESS) goto error;
   if((res = hal_param_float_new("quadspeed.setDY", HAL_RW, &parMain->DY, comp_id)) != HAL_SUCCESS) goto error;
   
   
   hal_ready(comp_id);
   return 0;

error:
    hal_exit(comp_id);
    return res;
}

void rtapi_app_exit(void) { hal_exit(comp_id); }
#endif
